#!/usr/bin/env python3

# Copyright (C) 2022-2023:
#   Helmholtz-Zentrum Potsdam Deutsches GeoForschungsZentrum GFZ
#
# This program is free software: you can redistribute it and/or modify it
# under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or (at
# your option) any later version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Affero
# General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program. If not, see http://www.gnu.org/licenses/.


import logging
import sys
import argparse
import configparser
import datetime
import csv

from exposurelib import SpatiaLiteExposure, PostGISExposure
from databaselib import PostGISDatabase
from taxonomylib import Taxonomy
from typing import Dict


# Add a logger printing error, warning, info and debug messages to the screen
logger = logging.getLogger()
logger.setLevel(logging.INFO)
logger.addHandler(logging.StreamHandler(sys.stdout))


class ExposureFinalizer:
    """
    This class creates the final exposure entities for a given country. The final exposure
    entities are based on the entities of the baseline exposure model imported with the
    Initializer. The Finalizer combines the baseline exposure with buildings for computing
    the final exposure. It uses as database either a SpatiaLite database for local storage or a
    PostGIS database for server-based storage.

    Args:
        exposure_database_object (AbstractExposure):
            Exposure database class as defined in the `exposure-lib`. Use either an instance of
            `SpatiaLiteExposure` for storing data in a local SpatiaLite database or
            `PostGISExposure` for storing data in a server-based PostGIS database.
        obm_database_object (AbstractExposure):
            OpenBuildingMap database class for access to building data. It is an instance of
            `PostGISDatabase` as defined in `database-lib`.
        occupancy_map_filepath (str):
            Filepath of the occupancy map (for simplifying occupancy types).
    """

    def __init__(
        self,
        exposure_database_object,
        obm_database_object,
        occupancy_map_filepath=None,
    ):
        self.exposure_db = exposure_database_object
        self.exposure_db.connect()
        self.obm_db = obm_database_object
        self.obm_db.connect()

        # Load the occupancy map
        self.occupancy_map = {}
        if occupancy_map_filepath is not None:
            self.occupancy_map = self.read_mapping_file(occupancy_map_filepath)

    @staticmethod
    def read_mapping_file(filepath: str) -> Dict[str, str]:
        """
        Reads a mapping table into a dictionary.

        Args:
            filepath (str):
                Filepath of the mapping table.
        """

        with open(filepath, mode="r") as input_file:
            reader = csv.reader(input_file)
            next(reader)
            return {rows[0]: rows[1] for rows in reader}

    def retrieve_buildings(self, country_iso_code):
        """
        Retrieves all buildings within the given country from the server-based building
        database and stores them in the local SpatiaLite database. To account for buildings
        outside the country but on tiles whose centroid is in the country, the country boundary
        is additionally buffered by 0.005 degrees (approx. 500m), resulting in a little more
        buildings being imported than truly needed. This function is useful only for the
        `local` runs of the Finalizer.

        Args:
            country_iso_code (str):
                ISO 3166-1 alpha-3 code of the country
        """

        # Retrieve country boundary with 500m buffer to include buildings located outside the
        # country but inside a tile whose centroid is inside the country
        logger.info("Retrieving country boundary with a buffer")
        boundary_geometry = self.exposure_db.get_country_boundary_geometry(
            country_iso_code, buffer=500
        )

        # Query buildings within the country boundary (with buffer)
        logger.info("Retrieving buildings within the country boundary")
        sql_statement = f"""
            SELECT osm_id, quadkey, occupancy, storeys, ST_AsText(geometry)
            FROM obm_buildings
            WHERE ST_Contains(
                ST_GeomFromText('{boundary_geometry}', 4326),
                ST_Centroid(obm_buildings.geometry)
                )
            """
        self.obm_db.cursor.execute(sql_statement)

        # Insert all queried buildings into the local database
        for num_buildings, (osm_id, quadkey, occupancy, stories, geom_wkt) in enumerate(
            self.obm_db.cursor
        ):
            self.exposure_db.insert_building(
                osm_id,
                quadkey,
                geom_wkt,
                occupancy=occupancy,
                stories=stories,
            )
            if (num_buildings + 1) % 10000 == 0:
                logger.info(f"{num_buildings + 1} buildings retrieved")
        logger.info("All buildings retrieved")

    def get_all_country_iso_codes(self):
        """
        Retrieves a list of all country ISO codes of all tiles in the database.

        Returns:
            List of country ISO codes.
        """

        sql_statement = f"SELECT DISTINCT country_iso_code FROM {self.exposure_db.tile_view}"
        self.exposure_db.cursor.execute(sql_statement)
        return self.exposure_db.cursor.fetchall()

    def process_country(self, country_iso_code):
        """
        Create the final exposure values (entities and assets) for the given country. This is
        done by first selecting all tiles of the country and then processing each tile in the
        function `process_tile`.

        Args:
            country_iso_code (str):
                ISO 3166-1 alpha-3 code of the country
        """

        # Retrieve the country assets as fallback set of assets
        sql_statement = f"""
            SELECT taxonomy_id, taxonomy_string, number_normalized,
                structural_normalized, night_normalized
            FROM AssetCountry
            INNER JOIN Taxonomy ON AssetCountry.taxonomy_id = Taxonomy.id
            WHERE country_iso_code = '{country_iso_code}'
            """
        self.exposure_db.cursor.execute(sql_statement)
        country_assets = self.exposure_db.cursor.fetchall()

        # Get all tiles within the country
        logger.info("Retrieving all tiles of country %s" % country_iso_code)
        if isinstance(self.exposure_db, SpatiaLiteExposure):
            y_max_function = "ST_MaxY"
        else:
            y_max_function = "ST_YMax"
        sql_statement = f"""
            SELECT quadkey, completeness
            FROM {self.exposure_db.tile_view}
            WHERE country_iso_code = '{country_iso_code}' AND built_area_size IS NOT NULL
            ORDER BY {y_max_function}({self.exposure_db.geometry_field}) DESC
            """
        self.exposure_db.cursor.execute(sql_statement)
        tiles = self.exposure_db.cursor.fetchall()

        # Process all tiles within the country
        logger.info(f"Processing all tiles of country {country_iso_code}")
        for idx, (quadkey, completeness) in enumerate(tiles):
            self.process_tile(quadkey, completeness, country_assets)
            if (idx + 1) % 1000 == 0:
                logger.info(f"Processed {idx + 1} of {len(tiles)} tiles")
            if (idx + 1) % 10000 == 0:
                self.exposure_db.connection.commit()
        logger.info(f"All {len(tiles)} tiles processed")
        self.exposure_db.connection.commit()

    def process_tile(self, quadkey, completeness, country_assets):
        """
        Creates the final exposure for the given tile with these processing steps:
        1. Copy the reference exposure data into the final data if existing
        2. Iterate through all buildings within the tile and
            - Assume the building to be represented by the reference assets (if exist) or by
              country assets (if reference assets do not exist)
            - Determine occupancy class and reduce asset list accordingly
            - Determine number of stories and reduce asset list accordingly
            - If no asset is left, return to starting set of assets
            - Normalize building numbers in assets to represent one building in total
            - Scale the tile assets down by one building
        3. If the tile is considered complete or the building number of assets is 0, remove all
           assets from the tile

        Args:
            quadkey (str):
                Quadkey of the tile.
            completeness (bool):
                Completeness of the tile.
            country_assets (list):
                List of country assets.
        """

        # Retrieve reference entity if it exists for the tile
        sql_statement = f"""
                SELECT id
                FROM EntityReference
                WHERE quadkey = '{quadkey}'
                """
        self.exposure_db.cursor.execute(sql_statement)
        reference_entity = self.exposure_db.cursor.fetchone()
        use_reference_entity = reference_entity is not None

        # Copy reference entity and reference assets to new tile entity
        # and create the background exposure model without buildings
        if use_reference_entity:
            tile_entity_id = self.exposure_db.insert_tile_entity(quadkey)
            sql_statement = f"""
                INSERT INTO Asset
                (entity_id, taxonomy_id, number, structural, night)
                SELECT {tile_entity_id}, taxonomy_id, number, structural, night
                FROM AssetReference WHERE entity_id = {reference_entity[0]}
                """
            self.exposure_db.cursor.execute(sql_statement)

        # Retrieve all buildings within the tile
        sql_statement = f"""
            SELECT osm_id, occupancy, storeys, ST_AsText({self.exposure_db.geometry_field})
            FROM {self.exposure_db.building_table}
            WHERE quadkey = '{quadkey}'
            """
        if isinstance(self.exposure_db, SpatiaLiteExposure):
            self.exposure_db.cursor.execute(sql_statement)
            buildings = self.exposure_db.cursor.fetchall()
        else:
            self.obm_db.cursor.execute(sql_statement)
            buildings = self.obm_db.cursor.fetchall()

        # If no buildings are present in the tile, processing can be finished here
        if len(buildings) == 0:
            return

        # Process all buildings in the tile
        if use_reference_entity:
            sql_statement = f"""
                SELECT taxonomy_id, taxonomy_string, number, structural, night
                FROM AssetReference
                INNER JOIN Taxonomy ON AssetReference.taxonomy_id = Taxonomy.id
                WHERE entity_id = {reference_entity[0]}
                """
            self.exposure_db.cursor.execute(sql_statement)
            tile_assets = self.exposure_db.cursor.fetchall()
            baseline_assets = tile_assets.copy()
        else:
            baseline_assets = country_assets.copy()

        tile_entity_id = self.exposure_db.insert_tile_entity(quadkey)
        for osm_id, occupancy, stories, geom_wkt in buildings:
            # Create building entity
            building_entity_id, _ = self.exposure_db.insert_building_entity(
                quadkey, osm_id, geom_wkt
            )
            # Determine the occupancy class
            if occupancy is not None:
                simple_occupancy = self.occupancy_map.get(occupancy, "PROBLEM")
                if simple_occupancy == "PROBLEM":
                    logger.warning("Unknown occupancy type detected: %s" % occupancy)
                    simple_occupancy = "UNK"
            else:
                simple_occupancy = "UNK"

            # Set the assets of the building
            assets = baseline_assets.copy()
            # Reduce assets of the building according to occupancy and number of stories
            assets = self.filter_assets(assets, simple_occupancy, stories)
            # If all assets are filtered out, try filtering with the country reference
            if len(assets) == 0:
                if use_reference_entity:
                    assets = country_assets.copy()
                    assets = self.filter_assets(assets, simple_occupancy, stories)
                    # If again all assets are filtered out, restoring to country assets
                    if len(assets) == 0:
                        assets = country_assets.copy()
                else:  # Country reference assets already used, therefore restoring them
                    assets = country_assets.copy()

            # Compute number of buildings from the set of assets for normalization
            sum_number = sum(map(lambda asset: asset[2], assets))

            # Normalize the asset list and insert it into the database
            normalized_assets = [
                [
                    building_entity_id,
                    taxonomy_id,
                    number / sum_number,
                    structural / sum_number,
                    night / sum_number,
                ]
                for taxonomy_id, _, number, structural, night in assets
            ]
            self.exposure_db.insert_assets(normalized_assets)

        # Reduce the tile assets by the number of buildings in the tile
        sql_statement = f"""
            SELECT SUM(Asset.number)
            FROM Asset
            INNER JOIN Entity ON Entity.id = Asset.entity_id
            WHERE Entity.quadkey = '{quadkey}' AND Entity.osm_id IS NULL
            """
        self.exposure_db.cursor.execute(sql_statement)
        number_assets = self.exposure_db.cursor.fetchone()

        # In case of a tile where tile assets exist
        if number_assets[0] is not None:
            # If more buildings were added to the tile than predicted by the model,
            # remove all assets from the tile entity
            if completeness or (number_assets[0] < len(buildings)):
                self.exposure_db.delete_assets(tile_entity_id)
            else:
                factor = (number_assets[0] - len(buildings)) / number_assets[0]
                sql_statement = f"""
                    UPDATE Asset
                    SET number = number * {factor},
                        structural = structural * {factor},
                        night = night * {factor}
                    WHERE entity_id = {tile_entity_id}
                    """
                self.exposure_db.cursor.execute(sql_statement)

    def filter_assets(self, assets, occupancy, stories):
        """
        Filter a list of assets based on a given occupancy and the number of stories. Only the
        assets that match the occupancy type and the number of stories are kept.

        Args:
            assets (list):
                List of assets to be filtered.
            occupancy (str):
                Occupancy code as filter criterion.
            stories (int):
                Number of stories as filter criterion.

        Returns:
            List of filtered assets.
        """

        # Reduce assets of the building according to occupancy
        if occupancy != "UNK":
            assets = list(
                filter(
                    lambda asset: (
                        self.get_taxonomy_section(asset[1], "occupancy") == occupancy
                    ),
                    assets,
                )
            )
        # Reduce assets of the building according to the number of stories
        if stories is not None:
            assets = list(
                filter(
                    lambda asset: (
                        self.check_stories(
                            self.get_taxonomy_section(asset[1], "height"), stories
                        )
                    ),
                    assets,
                )
            )
        return assets

    @staticmethod
    def get_taxonomy_section(taxonomy_string, section):
        """
        Retrieves the content of a section in a taxonomy string.

        Args:
            taxonomy_string (str):
                Taxonomy string from which the section content will be returned.
            section (str):
                Name of the section.

        Returns:
            Content of the section.
        """

        taxonomy = Taxonomy(taxonomy_string)
        return taxonomy.get_section(section)

    @staticmethod
    def check_stories(taxonomy_part, number_stories):
        """
        Checks if the given number of stories of a building are matching the content of the
        height section of the taxonomy string. Two taxonomy codes are considered here: `H` and
        `HBET`.

        Args:
            taxonomy_part (str):
                The `height` section of the building's taxonomy string
            number_stories (int):
                Number of stories of the building

        Returns:
            True if the number of stories match the content of the height section of the
            taxonomy string.
        """

        # Raise an error if the `height` section is `None` to avoid returning `False`
        if taxonomy_part is None:
            raise ValueError("check_stories: `taxonomy_part` cannot be `None`")
        # If the `height` section in the taxonomy is empty, keep the taxonomy
        if taxonomy_part == "":
            return True
        # Analyze the tags in the `height` section
        key, value = taxonomy_part.split(":")
        if key == "HBET":
            values = value.split("-")
            if values[0] == "":
                min_stories = 0
            else:
                min_stories = int(values[0])
            if values[1] == "":
                max_stories = 10000
            else:
                max_stories = int(values[1])
            if (number_stories >= min_stories) and (number_stories <= max_stories):
                return True
        if key == "H":
            if int(value) == number_stories:
                return True
        return False


def command_line_interface():
    """
    Command-line interface of the exposure-finalizer. The first command-line option is the
    command defining what type of database will be used:
        `local` : Exposure will be generated in a local SpatiaLite database.
        `server`: Exposure will be generated in a server-based PostGIS database.
    The commands require these mandatory options:
        `-e`: Filepath of the exposure database (`local` only).
        `-c`: Config file (INI-type) describing the access to the tile database. Also needed
              for `local` as the boundaries and tiles will be copied over to the local
              SpatiaLite database.
        `-o`: Filepath of the occupancy map (for simplifying occupancy types).
    The `server` command requires the following option:
        `-i`: ISO 3166-1 alpha-3 code of the country for which the exposure is to be imported.
    The `-i` option is optional for the `local` command. If given, buildings for the given
    country ISO code will be imported and all tiles of that country will be processed. If not
    given, Finalizer will process all tiles in the local database assuming that all buildings
    are present (such a file can be created with exposure-share using the `baseline` command
    and the `-b` option).
    """

    # Create the argument parser
    parser = argparse.ArgumentParser(description="Exposure-Finalizer")
    parser_shared = argparse.ArgumentParser(add_help=False)
    parser_shared.add_argument(
        "-c",
        "--config-file",
        required=True,
        type=str,
        help="Filepath of the config file for accessing the Tiles database",
    )
    parser_shared.add_argument(
        "-o",
        "--occupancy-map",
        required=False,
        type=str,
        help="Filepath to the occupancy-map file",
    )

    # Prepare for subparsers
    subparsers = parser.add_subparsers(help="sub-command help", dest="command")
    # Create the parser for the 'local' command
    parser_local = subparsers.add_parser(
        "local", help="Run builder on a local SpatiaLite database", parents=[parser_shared]
    )
    parser_local.add_argument(
        "-e",
        "--exposure-database",
        required=True,
        type=str,
        help="Filepath to the SpatiaLite exposure database",
    )
    parser_local.add_argument(
        "-i",
        "--country-iso-code",
        required=False,
        type=str,
        help="ISO 3166-1 alpha-3 code of the country",
    )
    # Create the parser for the 'server' command
    parser_server = subparsers.add_parser(
        "server", help="Run builder on a server-based PostGIS database", parents=[parser_shared]
    )
    parser_server.add_argument(
        "-i",
        "--country-iso-code",
        required=True,
        type=str,
        help="ISO 3166-1 alpha-3 code of the country",
    )

    # read arguments from command line.
    args = parser.parse_args()
    command = args.command
    occupancy_map_filepath = args.occupancy_map
    config_file = args.config_file
    country_iso_code = args.country_iso_code

    postgis_config = configparser.ConfigParser()
    postgis_config.read(config_file)

    # Initialize database
    exposure_database_object = None
    if command == "local":
        exposure_database_object = SpatiaLiteExposure(args.exposure_database)
    elif command == "server":
        exposure_db_config = {
            "host": postgis_config["Exposure"]["host"],
            "dbname": postgis_config["Exposure"]["dbname"],
            "port": postgis_config["Exposure"]["port"],
            "username": postgis_config["Exposure"]["username"],
            "password": postgis_config["Exposure"]["password"],
        }
        exposure_database_object = PostGISExposure(**exposure_db_config)

    obm_db_config = {
        "host": postgis_config["OBM"]["host"],
        "dbname": postgis_config["OBM"]["dbname"],
        "port": postgis_config["OBM"]["port"],
        "username": postgis_config["OBM"]["username"],
        "password": postgis_config["OBM"]["password"],
    }
    obm_database_object = PostGISDatabase(**obm_db_config)

    start_time = datetime.datetime.now()
    logger.info("Start time: %s" % start_time)

    if exposure_database_object is not None:
        finalizer = ExposureFinalizer(
            exposure_database_object,
            obm_database_object,
            occupancy_map_filepath,
        )
        # If ISO is not given, assume the buildings are present in the database
        if (command == "local") and (country_iso_code is not None):
            finalizer.retrieve_buildings(country_iso_code)

        # If all tiles are processed, identify the country ISO codes of them
        if (command == "local") and (country_iso_code is None):
            logger.debug("Retrieving all country ISO codes")
            iso_codes = finalizer.get_all_country_iso_codes()
        else:
            # Set the given ISO code for processing
            iso_codes = [(country_iso_code,)]
        # Process all given countries sequentially
        for iso_code, *_ in iso_codes:
            logger.info(f"Processing country {iso_code}")
            finalizer.exposure_db.remove_exposure(iso_code, delete_reference=False)
            finalizer.process_country(iso_code)

        finalizer.exposure_db.commit_and_close()
    else:
        logger.info("No database connection available")

    logger.info("Execution time: %s" % (datetime.datetime.now() - start_time))


if __name__ == "__main__":
    command_line_interface()
